use super::*;
use std::io;
use cryptohelpers::sha256;
use cryptohelpers::sha2::{
    Digest, Sha256,
};
use tokio_uring::fs::{
    File, OpenOptions,
};
use tokio::sync::{
    mpsc,
    oneshot,   
};
use tokio_stream::wrappers::ReceiverStream;
use futures::prelude::*;

async fn uring_read<F>(file: &mut File, mut to: F) -> io::Result<usize>
where F: FnMut(&[u8]) -> io::Result<()>
{
    let mut full_buffer = vec![0u8; 4096]; // we need to allocate this so &buffer[0] is always the same.

    let mut read = 0usize;
    loop {
	let buffer = {
	    let (res, n_full_buffer) = file.read_at(full_buffer, read as u64).await;

	    full_buffer = n_full_buffer;
	    &full_buffer[..(match res {
		Ok(0) => break,
		Ok(n) => n,
		Err(e) => return Err(e),
	    })]
	};
	to(buffer)?;
	read += buffer.len();
    }
    Ok(read)
}

fn file_handler(mut recv: mpsc::Receiver<PathBuf>, backpressure: usize) -> impl Stream<Item = (PathBuf, io::Result<sha256::Sha256Hash>)>
{
    
    let (r_tx, r_rx) = mpsc::channel(backpressure);
    std::thread::spawn(move || {
	tokio_uring::start(async move {
	    
	    while let Some(path) = recv.recv().await {
		let ret = r_tx.clone();
		tokio_uring::spawn(async move {
		    let mut file = OpenOptions::new()
			.read(true)
			.open(&path).await.unwrap();
		    let mut hasher = Sha256::new();
		    let ring_res = uring_read(&mut file, |buffer| {
			if ret.is_closed() {
			    return Err(io::Error::new(io::ErrorKind::ConnectionAborted, "return channel dropped, stopping operation"));
			}
			hasher.update(buffer);
			Ok(())
		    }).await;
		    let _ = tokio::join![
			file.close(), // We are in a unique task per file, so awaiting this here concurrently with the returning async block is fine.
			async move {
			    match ring_res {
				Ok(_n) => {
				    eprintln!("Proc {} bytes from {:?}", _n, path);
				    let _ = ret.send((path, Ok(hasher.into()))).await;
				},
				Err(e) => {
				    eprintln!("Proc for {:?} failed: {}", path, e);
				    match ret.try_send((path, Err(e))) {
					Ok(_) => (),
					Err(mpsc::error::TrySendError::Full(val)) => { tokio_uring::spawn(async move { let _ = ret.send(val).await; }); },
					_ => return,
				    }
				}
			    }
			}
		    ];
		});
	    }
	    // Drop sender that we're cloning from
	    drop(r_tx);
	});
    });
    ReceiverStream::new(r_rx)
}
//TODO: Higher-level wrapper around `file_handler()`
